/* GStreamer AAC+ (libaacplus Encoder) plugin
 * Copyright (C) 2011 Sergiy Guryev <piratfm@ua.fm>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-aacplus
 * @see_also: faac
 *
 * aacplus encodes raw audio to AAC (SBR or SBR+PS) (MPEG-4 part 3) streams.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch audiotestsrc wave=sine num-buffers=100 ! audioconvert ! aacplus ! matroskamux ! filesink location=sine.mkv
 * ]| Encode a sine beep as aac+ and write to matroska container.
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdlib.h>
#include <string.h>

#include <gst/audio/multichannel.h>
#include <gst/pbutils/codec-utils.h>

#include "gstaacplus.h"

#define SAMPLE_RATES "32000, " \
                    "44100, " \
                    "48000"
#define SAMPLE_RATES_AAC " 16000, " \
                    "22050, " \
                    "24000 "

#define SINK_CAPS \
    "audio/x-raw-int, "                \
    "endianness = (int) BYTE_ORDER, "  \
    "signed = (boolean) true, "        \
    "width = (int) 16, "               \
    "depth = (int) 16, "               \
    "rate = (int) {" SAMPLE_RATES "}, "   \
    "channels = (int) [ 1, 2 ] "

#define SRC_CAPS \
    "audio/mpeg, "                     \
    "mpegversion = (int) 4, "   \
    "channels = (int) [ 1, 2 ], "      \
    "rate = (int) {" SAMPLE_RATES_AAC "}, "   \
    "stream-format = (string) { adts, raw }, " \
    "base-profile = (string) { sbr, ps }; " \

#if 0
	"audio/mpeg, "                     \
    "mpegversion = (int) 2, "   \
    "channels = (int) [ 1, 2 ], "      \
    "rate = (int) {" SAMPLE_RATES "}, "   \
    "stream-format = (string) { adts, raw }, " \
    "profile = (string) { main, lc }"
#endif

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (SRC_CAPS));

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (SINK_CAPS));

enum
{
  PROP_0,
  PROP_BITRATE,
};

static void gst_aacplus_base_init (GstAacPlusClass * klass);
static void gst_aacplus_class_init (GstAacPlusClass * klass);
static void gst_aacplus_init (GstAacPlus * aacplus);
static void gst_aacplus_finalize (GObject * object);
static void gst_aacplus_reset (GstAacPlus * aacplus);

static void gst_aacplus_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec);
static void gst_aacplus_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec);

static gboolean gst_aacplus_sink_event (GstPad * pad, GstEvent * event);
static gboolean gst_aacplus_configure_source_pad (GstAacPlus * aacplus);
static gboolean gst_aacplus_sink_setcaps (GstPad * pad, GstCaps * caps);
static GstCaps *gst_aacplus_sink_getcaps (GstPad * pad);
static GstFlowReturn gst_aacplus_push_buffers (GstAacPlus * aacplus, gboolean force);
static GstFlowReturn gst_aacplus_chain (GstPad * pad, GstBuffer * data);
static GstStateChangeReturn gst_aacplus_change_state (GstElement * element,
    GstStateChange transition);

static GstElementClass *parent_class = NULL;

GST_DEBUG_CATEGORY_STATIC (aacplus_debug);
#define GST_CAT_DEFAULT aacplus_debug

#define AACPLUS_DEFAULT_OUTPUTFORMAT 0     /* RAW */
#define AACPLUS_DEFAULT_BITRATE      40 * 1000

GType
gst_aacplus_get_type (void)
{
  static GType gst_aacplus_type = 0;

  if (!gst_aacplus_type) {
    static const GTypeInfo gst_aacplus_info = {
      sizeof (GstAacPlusClass),
      (GBaseInitFunc) gst_aacplus_base_init,
      NULL,
      (GClassInitFunc) gst_aacplus_class_init,
      NULL,
      NULL,
      sizeof (GstAacPlus),
      0,
      (GInstanceInitFunc) gst_aacplus_init,
    };
    const GInterfaceInfo preset_interface_info = {
      NULL,                     /* interface_init */
      NULL,                     /* interface_finalize */
      NULL                      /* interface_data */
    };

    gst_aacplus_type = g_type_register_static (GST_TYPE_ELEMENT,
        "GstAacPlus", &gst_aacplus_info, 0);

    g_type_add_interface_static (gst_aacplus_type, GST_TYPE_PRESET,
        &preset_interface_info);
  }

  return gst_aacplus_type;
}

static void
gst_aacplus_base_init (GstAacPlusClass * klass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_template));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_template));

  gst_element_class_set_details_simple (element_class, "AAC+ audio encoder",
      "Codec/Encoder/Audio",
      "3GPP's MPEG-2/4 AAC+ encoder",
      "Sergiy Gur'yev <piratfm@ua.fm>");

  GST_DEBUG_CATEGORY_INIT (aacplus_debug, "aacplus", 0, "AAC+ encoding");
}


static void
gst_aacplus_class_init (GstAacPlusClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *gstelement_class = GST_ELEMENT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gst_aacplus_set_property;
  gobject_class->get_property = gst_aacplus_get_property;
  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_aacplus_finalize);

  /* properties */
  g_object_class_install_property (gobject_class, PROP_BITRATE,
      g_param_spec_int ("bitrate", "Bitrate (bps)",
          "Average bitrate (ABR) in bits/sec", 10 * 1000, 72 * 1000,
          AACPLUS_DEFAULT_BITRATE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  /* virtual functions */
  gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_aacplus_change_state);
}

static void
gst_aacplus_init (GstAacPlus * aacplus)
{
  aacplus->sinkpad = gst_pad_new_from_static_template (&sink_template, "sink");
  gst_pad_set_chain_function (aacplus->sinkpad,
      GST_DEBUG_FUNCPTR (gst_aacplus_chain));
  gst_pad_set_setcaps_function (aacplus->sinkpad,
      GST_DEBUG_FUNCPTR (gst_aacplus_sink_setcaps));
  gst_pad_set_getcaps_function (aacplus->sinkpad,
      GST_DEBUG_FUNCPTR (gst_aacplus_sink_getcaps));
  gst_pad_set_event_function (aacplus->sinkpad,
      GST_DEBUG_FUNCPTR (gst_aacplus_sink_event));
  gst_element_add_pad (GST_ELEMENT (aacplus), aacplus->sinkpad);

  aacplus->srcpad = gst_pad_new_from_static_template (&src_template, "src");
  gst_pad_use_fixed_caps (aacplus->srcpad);
  gst_element_add_pad (GST_ELEMENT (aacplus), aacplus->srcpad);

  aacplus->adapter = gst_adapter_new ();

  /* default properties */
  aacplus->bitrate = AACPLUS_DEFAULT_BITRATE;
  aacplus->outputformat = AACPLUS_DEFAULT_OUTPUTFORMAT;

  gst_aacplus_reset (aacplus);
}

static void
gst_aacplus_reset (GstAacPlus * aacplus)
{
  aacplus->handle = NULL;
  aacplus->samplerate = -1;
  aacplus->channels = -1;
  aacplus->offset = 0;
  gst_adapter_clear (aacplus->adapter);
}

static void
gst_aacplus_finalize (GObject * object)
{
  GstAacPlus *aacplus = (GstAacPlus *) object;

  g_object_unref (aacplus->adapter);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
gst_aacplus_close_encoder (GstAacPlus * aacplus)
{
  if (aacplus->handle)
    aacplusEncClose (aacplus->handle);
  aacplus->handle = NULL;
  gst_adapter_clear (aacplus->adapter);
  aacplus->offset = 0;
}

static const GstAudioChannelPosition aac_channel_positions[][2] = {
  {GST_AUDIO_CHANNEL_POSITION_FRONT_MONO},
  {GST_AUDIO_CHANNEL_POSITION_FRONT_LEFT,
      GST_AUDIO_CHANNEL_POSITION_FRONT_RIGHT},
};

static GstCaps *
gst_aacplus_sink_getcaps (GstPad * pad)
{
  static volatile gsize sinkcaps = 0;

  if (g_once_init_enter (&sinkcaps)) {
    GstCaps *tmp = gst_caps_new_empty ();
    GstStructure *s, *t;
    gint i, c;
    static const int rates[] = {
      32000, 44100, 48000
    };
    GValue rates_arr = { 0, };
    GValue tmp_v = { 0, };

    g_value_init (&rates_arr, GST_TYPE_LIST);
    g_value_init (&tmp_v, G_TYPE_INT);
    for (i = 0; i < G_N_ELEMENTS (rates); i++) {
      g_value_set_int (&tmp_v, rates[i]);
      gst_value_list_append_value (&rates_arr, &tmp_v);
    }
    g_value_unset (&tmp_v);

    s = gst_structure_new ("audio/x-raw-int",
        "endianness", G_TYPE_INT, G_BYTE_ORDER,
        "signed", G_TYPE_BOOLEAN, TRUE,
        "width", G_TYPE_INT, 16, "depth", G_TYPE_INT, 16, NULL);
    gst_structure_set_value (s, "rate", &rates_arr);

    for (i = 1; i <= 2; i++) {
      GValue chanpos = { 0 };
      GValue pos = { 0 };

      t = gst_structure_copy (s);

      gst_structure_set (t, "channels", G_TYPE_INT, i, NULL);

      g_value_init (&chanpos, GST_TYPE_ARRAY);
      g_value_init (&pos, GST_TYPE_AUDIO_CHANNEL_POSITION);

      for (c = 0; c < i; c++) {
        g_value_set_enum (&pos, aac_channel_positions[i - 1][c]);
        gst_value_array_append_value (&chanpos, &pos);
      }
      g_value_unset (&pos);

      gst_structure_set_value (t, "channel-positions", &chanpos);
      g_value_unset (&chanpos);
      gst_caps_append_structure (tmp, t);
    }
    gst_structure_free (s);
    g_value_unset (&rates_arr);

    GST_DEBUG_OBJECT (pad, "Generated sinkcaps: %" GST_PTR_FORMAT, tmp);

    g_once_init_leave (&sinkcaps, (gsize) tmp);
  }

  return gst_caps_ref ((GstCaps *) sinkcaps);
}

/* check downstream caps to configure format */
static void
gst_aacplus_negotiate (GstAacPlus * aacplus)
{
  GstCaps *caps;

  caps = gst_pad_get_allowed_caps (aacplus->srcpad);

  GST_DEBUG_OBJECT (aacplus, "allowed caps: %" GST_PTR_FORMAT, caps);

  if (caps && gst_caps_get_size (caps) > 0) {
    GstStructure *s = gst_caps_get_structure (caps, 0);
    const gchar *str = NULL;

    if ((str = gst_structure_get_string (s, "stream-format"))) {
      if (strcmp (str, "adts") == 0) {
        GST_DEBUG_OBJECT (aacplus, "use ADTS format for output");
        aacplus->outputformat = 1;
      } else if (strcmp (str, "raw") == 0) {
        GST_DEBUG_OBJECT (aacplus, "use RAW format for output");
        aacplus->outputformat = 0;
      } else {
        GST_DEBUG_OBJECT (aacplus, "unknown stream-format: %s", str);
        aacplus->outputformat = 0;
      }
    }
  }

  if (caps)
    gst_caps_unref (caps);
}

static gboolean
gst_aacplus_sink_setcaps (GstPad * pad, GstCaps * caps)
{
  GstAacPlus *aacplus = GST_AACPLUS (gst_pad_get_parent (pad));
  GstStructure *structure = gst_caps_get_structure (caps, 0);
  aacplusEncHandle *handle;
  gint channels, samplerate, width;
  gulong samples, bytes, bps = 0;
  gint fmt = -1;
  gboolean result = FALSE;

  if (!gst_caps_is_fixed (caps))
    goto refuse_caps;

  if (!gst_structure_get_int (structure, "channels", &channels) ||
      !gst_structure_get_int (structure, "rate", &samplerate)) {
    goto refuse_caps;
  }

  if (gst_structure_has_name (structure, "audio/x-raw-int")) {
    gst_structure_get_int (structure, "width", &width);
    switch (width) {
      case 16:
        fmt = AACPLUS_INPUT_16BIT;
        bps = 2;
        break;
      default:
        g_return_val_if_reached (FALSE);
    }
  } else {
    fmt = AACPLUS_INPUT_FLOAT;
    bps = 4;
  }

  if (fmt == -1)
    goto refuse_caps;

  /* If the encoder is initialized, do not
     reinitialize it again if not necessary */
  if (aacplus->handle) {
    if (samplerate == aacplus->samplerate && channels == aacplus->channels &&
        fmt == aacplus->format)
      return TRUE;

    /* clear out pending frames */
    gst_aacplus_push_buffers (aacplus, TRUE);

    gst_aacplus_close_encoder (aacplus);
  }

  if (!(handle = aacplusEncOpen (samplerate, channels, &samples, &bytes)))
    goto setup_failed;

  /* ok, record and set up */
  aacplus->format = fmt;
  aacplus->bps = bps;
  aacplus->handle = handle;
  aacplus->bytes = bytes;
  aacplus->samples = samples;
  aacplus->channels = channels;
  aacplus->samplerate = samplerate;

  gst_aacplus_negotiate (aacplus);

  /* finish up */
  result = gst_aacplus_configure_source_pad (aacplus);

done:
  gst_object_unref (aacplus);
  return result;

  /* ERRORS */
setup_failed:
  {
    GST_ELEMENT_ERROR (aacplus, LIBRARY, SETTINGS, (NULL), (NULL));
    goto done;
  }
refuse_caps:
  {
    GST_WARNING_OBJECT (aacplus, "refused caps %" GST_PTR_FORMAT, caps);
    goto done;
  }
}

static gboolean
gst_aacplus_configure_source_pad (GstAacPlus * aacplus)
{
  GstCaps *srccaps;
  gboolean ret = FALSE;
  aacplusEncConfiguration *conf;
  guint maxbitrate;

  /* we negotiated caps update current configuration */
  conf = aacplusEncGetCurrentConfiguration (aacplus->handle);
  conf->bitRate = aacplus->bitrate;

  conf->inputFormat = aacplus->format;
  conf->outputFormat = aacplus->outputformat;

  /* check, warn and correct if the max bitrate for the given samplerate is
   * exceeded. Maximum of 6144 bit for a channel */
  maxbitrate =
      (unsigned int) (6144.0 * (double) aacplus->samplerate / (double) 1024.0 +
      .5);
  if (conf->bitRate > maxbitrate) {
    GST_ELEMENT_WARNING (aacplus, RESOURCE, SETTINGS, (NULL),
        ("bitrate %lu exceeds maximum allowed bitrate of %u for samplerate %d. "
            "Setting bitrate to %u", conf->bitRate, maxbitrate,
            aacplus->samplerate, maxbitrate));
    conf->bitRate = maxbitrate;
  }

  /* default 0 to start with, libaacplus chooses based on bitrate */
  conf->bandWidth = 0;

  if (!aacplusEncSetConfiguration (aacplus->handle, conf))
    goto set_failed;

  /* let's see what really happened,
   * note that this may not really match desired rate */
  GST_DEBUG_OBJECT (aacplus, "bitrate: %lu kbps", conf->bitRate / 1000);
  GST_DEBUG_OBJECT (aacplus, "bandwidth: %d Hz", conf->bandWidth);
  GST_DEBUG_OBJECT (aacplus, "samplerate transformation: %d => %d Hz",
		  aacplus->samplerate, conf->sampleRate);
  GST_DEBUG_OBJECT (aacplus, "channels transformation: %d => %d",
		  aacplus->channels, conf->nChannelsOut);

  /* now create a caps for it all */
  srccaps = gst_caps_new_simple ("audio/mpeg",
      "mpegversion", G_TYPE_INT, 4,
      "channels", G_TYPE_INT, conf->nChannelsOut,
      "rate", G_TYPE_INT, conf->sampleRate,
      "stream-format", G_TYPE_STRING, (aacplus->outputformat ? "adts" : "raw"),
      NULL);

  guint8 *config = NULL;
  gulong config_len = 0;

  /* get the config string */
  GST_DEBUG_OBJECT (aacplus, "retrieving decoder info");
  aacplusEncGetDecoderSpecificInfo (aacplus->handle, &config, &config_len);

  if (!gst_codec_utils_aac_caps_set_level_and_profile (srccaps, config,
		  config_len)) {
	free (config);
	gst_caps_unref (srccaps);
	goto invalid_codec_data;
  }

  if (!aacplus->outputformat) {
	GstBuffer *codec_data;

	/* copy it into a buffer */
	codec_data = gst_buffer_new_and_alloc (config_len);
	memcpy (GST_BUFFER_DATA (codec_data), config, config_len);

	/* add to caps */
	gst_caps_set_simple (srccaps,
	  "codec_data", GST_TYPE_BUFFER, codec_data, NULL);

	gst_buffer_unref (codec_data);
  }

  free (config);

  GST_DEBUG_OBJECT (aacplus, "src pad caps: %" GST_PTR_FORMAT, srccaps);

  ret = gst_pad_set_caps (aacplus->srcpad, srccaps);
  gst_caps_unref (srccaps);

  return ret;

  /* ERROR */
set_failed:
  {
    GST_WARNING_OBJECT (aacplus, "AacPlus doesn't support the current configuration");
    return FALSE;
  }
invalid_codec_data:
  {
    GST_ERROR_OBJECT (aacplus, "Invalid codec data");
    return FALSE;
  }
}

static GstFlowReturn
gst_aacplus_push_buffers (GstAacPlus * aacplus, gboolean force)
{
  GstFlowReturn ret = GST_FLOW_OK;
  gint av, frame_size, size, ret_size;
  GstBuffer *outbuf;
  guint64 timestamp, distance;
  const guint8 *data;

  /* samples already considers channel count */
  frame_size = aacplus->samples * aacplus->bps;

  while (G_LIKELY (ret == GST_FLOW_OK)) {

    av = gst_adapter_available (aacplus->adapter);

    GST_LOG_OBJECT (aacplus, "pushing: force: %d, frame_size: %d, av: %d, "
        "offset: %d", force, frame_size, av, aacplus->offset);

    /* idea:
     * - start of adapter corresponds with what has already been encoded
     * (i.e. really returned by aacplus)
     * - start + offset is what needs to be fed to aacplus next
     * That way we can timestamp the output based
     * on adapter provided timestamp (and duration is a fixed frame duration) */

    /* not enough data for one frame and no flush forcing */
    if (!force && (av < frame_size + aacplus->offset))
      break;

    if (G_LIKELY (av - aacplus->offset >= frame_size)) {
      GST_LOG_OBJECT (aacplus, "encoding a frame");
      data = gst_adapter_peek (aacplus->adapter, aacplus->offset + frame_size);
      data += aacplus->offset;
      size = frame_size;
    } else if (av - aacplus->offset > 0) {
      GST_LOG_OBJECT (aacplus, "encoding leftover");
      data = gst_adapter_peek (aacplus->adapter, av);
      data += aacplus->offset;
      size = av - aacplus->offset;
    } else {
      GST_LOG_OBJECT (aacplus, "emptying encoder");
      data = NULL;
      size = 0;
    }

    outbuf = gst_buffer_new_and_alloc (aacplus->bytes);

    if (G_UNLIKELY ((ret_size = aacplusEncEncode (aacplus->handle, (gint32 *) data,
                    size / aacplus->bps, GST_BUFFER_DATA (outbuf),
                    aacplus->bytes)) < 0)) {
      gst_buffer_unref (outbuf);
      goto encode_failed;
    }

    GST_LOG_OBJECT (aacplus, "encoder return: %d", ret_size);

    /* consumed, advanced view */
    aacplus->offset += size;
    g_assert (aacplus->offset <= av);

    if (G_UNLIKELY (!ret_size)) {
      gst_buffer_unref (outbuf);
      if (size)
        continue;
      else
        break;
    }

    /* deal with encoder lead-out */
    if (G_UNLIKELY (av == 0 && aacplus->offset == 0)) {
      GST_DEBUG_OBJECT (aacplus, "encoder returned additional data");
      /* continuous with previous output, ok to have 0 duration */
      timestamp = aacplus->next_ts;
    } else {
      /* after some caching, finally some data */
      /* adapter gives time */
      timestamp = gst_adapter_prev_timestamp (aacplus->adapter, &distance);
    }

    if (G_LIKELY ((av = gst_adapter_available (aacplus->adapter)) >= frame_size)) {
      /* must have then come from a complete frame */
      gst_adapter_flush (aacplus->adapter, frame_size);
      aacplus->offset -= frame_size;
      size = frame_size;
    } else {
      /* otherwise leftover */
      gst_adapter_clear (aacplus->adapter);
      aacplus->offset = 0;
      size = av;
    }

    GST_BUFFER_SIZE (outbuf) = ret_size;
    if (G_LIKELY (GST_CLOCK_TIME_IS_VALID (timestamp)))
      GST_BUFFER_TIMESTAMP (outbuf) = timestamp +
          GST_FRAMES_TO_CLOCK_TIME (distance / aacplus->channels / aacplus->bps,
          aacplus->samplerate);
    GST_BUFFER_DURATION (outbuf) =
        GST_FRAMES_TO_CLOCK_TIME (size / aacplus->channels / aacplus->bps,
        aacplus->samplerate);
    aacplus->next_ts =
        GST_BUFFER_TIMESTAMP (outbuf) + GST_BUFFER_DURATION (outbuf);

    /* perhaps check/set DISCONT based on timestamps ? */

    GST_LOG_OBJECT (aacplus, "Pushing out buffer time: %" GST_TIME_FORMAT
        " duration: %" GST_TIME_FORMAT,
        GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (outbuf)),
        GST_TIME_ARGS (GST_BUFFER_DURATION (outbuf)));

    gst_buffer_set_caps (outbuf, GST_PAD_CAPS (aacplus->srcpad));
    ret = gst_pad_push (aacplus->srcpad, outbuf);
  }

  /* in case encoder returns less than expected, clear our view as well */
  if (G_UNLIKELY (force)) {
#ifndef GST_DISABLE_GST_DEBUG
    if ((av = gst_adapter_available (aacplus->adapter)))
      GST_WARNING_OBJECT (aacplus, "encoder left %d bytes; discarding", av);
#endif
    gst_adapter_clear (aacplus->adapter);
    aacplus->offset = 0;
  }

  return ret;

  /* ERRORS */
encode_failed:
  {
    GST_ELEMENT_ERROR (aacplus, LIBRARY, ENCODE, (NULL), (NULL));
    return GST_FLOW_ERROR;
  }
}

static gboolean
gst_aacplus_sink_event (GstPad * pad, GstEvent * event)
{
  GstAacPlus *aacplus;
  gboolean ret;

  aacplus = GST_AACPLUS (gst_pad_get_parent (pad));

  GST_LOG_OBJECT (aacplus, "received %s", GST_EVENT_TYPE_NAME (event));

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_EOS:
    {
      if (aacplus->handle) {
        /* flush first */
        GST_DEBUG_OBJECT (aacplus, "Pushing out remaining buffers because of EOS");
        gst_aacplus_push_buffers (aacplus, TRUE);
      }

      ret = gst_pad_event_default (pad, event);
      break;
    }
    default:
      ret = gst_pad_event_default (pad, event);
      break;

  }
  gst_object_unref (aacplus);
  return ret;
}

static GstFlowReturn
gst_aacplus_chain (GstPad * pad, GstBuffer * inbuf)
{
  GstFlowReturn result = GST_FLOW_OK;
  GstAacPlus *aacplus;

  aacplus = GST_AACPLUS (gst_pad_get_parent (pad));

  if (!aacplus->handle)
    goto no_handle;

  GST_LOG_OBJECT (aacplus, "Got buffer time: %" GST_TIME_FORMAT " duration: %"
      GST_TIME_FORMAT, GST_TIME_ARGS (GST_BUFFER_TIMESTAMP (inbuf)),
      GST_TIME_ARGS (GST_BUFFER_DURATION (inbuf)));

  gst_adapter_push (aacplus->adapter, inbuf);

  result = gst_aacplus_push_buffers (aacplus, FALSE);

done:
  gst_object_unref (aacplus);

  return result;

  /* ERRORS */
no_handle:
  {
    GST_ELEMENT_ERROR (aacplus, CORE, NEGOTIATION, (NULL),
        ("format wasn't negotiated before chain function"));
    gst_buffer_unref (inbuf);
    result = GST_FLOW_ERROR;
    goto done;
  }
}

static void
gst_aacplus_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstAacPlus *aacplus = GST_AACPLUS (object);

  GST_OBJECT_LOCK (aacplus);

  switch (prop_id) {
    case PROP_BITRATE:
      aacplus->bitrate = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

  GST_OBJECT_UNLOCK (aacplus);
}

static void
gst_aacplus_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstAacPlus *aacplus = GST_AACPLUS (object);

  GST_OBJECT_LOCK (aacplus);

  switch (prop_id) {
    case PROP_BITRATE:
      g_value_set_int (value, aacplus->bitrate);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

  GST_OBJECT_UNLOCK (aacplus);
}

static GstStateChangeReturn
gst_aacplus_change_state (GstElement * element, GstStateChange transition)
{
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  GstAacPlus *aacplus = GST_AACPLUS (element);

  /* upwards state changes */
  switch (transition) {
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);

  /* downwards state changes */
  switch (transition) {
    case GST_STATE_CHANGE_PAUSED_TO_READY:
    {
      gst_aacplus_close_encoder (aacplus);
      gst_aacplus_reset (aacplus);
      break;
    }
    default:
      break;
  }

  return ret;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, "aacplus", GST_RANK_MARGINAL,
      GST_TYPE_AACPLUS);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "aacplus",
    "AAC+ encoder (libaacplus)",
    plugin_init, VERSION, "LGPL", "libaacplus", "http://tipok.org.ua/node/17")
